using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace ConvexHulls
{
    /// <summary>
    /// Convex hull algorithm "Graham scan".
    /// Web: http://de.wikipedia.org/wiki/Graham_Scan
    /// </summary>
    class GrahamScan : IConvexHull
    {
        /// <summary>
        /// Compares two points p1, p2 by means of a given point p0.
        /// If the angle from p0->p1 with the x axis is littler than that of p1->p2, Compare() returns
        /// a negative value. If the angles are equal, 0 is returned. Positive value else.
        /// </summary>
        private class AngleComparer : IComparer<Point>
        {
            private Point p0;

            public AngleComparer(Point p0)
            {
                this.p0 = p0;
            }

            public int Compare(Point p1, Point p2)
            {
                // compare angles
                int res = (p1.X - p0.X) * (p2.Y - p0.Y) - (p2.X - p0.X) * (p1.Y - p0.Y);
                // if same angles, then compare distances from p0
                if (res == 0 && dist(p1) < dist(p2))
                    res = -1;
                return res;
            }

            // Returns the "distance" of p to p0 by absolute differences.
            private int dist(Point p)
            {
                return Math.Abs(p.X - p0.X) + Math.Abs(p.Y - p0.Y);
            }
        }

        /// <summary>
        /// Calculates the convex hull.
        /// </summary>
        public Point[] CalculateHull(Point[] points)
        {
            if (points.Length <= 1)
                return points;

            // search for the lowest ordinate
            Point p0 = points[0];
            for (int ii = 1; ii < points.Length; ii++)
            {
                if (points[ii].Y < p0.Y || (points[ii].Y == p0.Y && points[ii].X < p0.X))
                    p0 = points[ii];
            }
            
            // create array of points without p0
            Point[] S = new Point[points.Length - 1];
            for (int ii = 0, j = 0; ii < points.Length; ii++)
            {
                if (points[ii] == p0)
                    j = 1;
                else
                    S[ii - j] = points[ii];
            }
            // sort this array with ascending angle
            Array.Sort<Point>(S, new AngleComparer(p0));

            // finally... calculate the hull
            Stack<Point> K = new Stack<Point>();
            K.Push(p0);
            K.Push(S[0]);
            int i = 1;
            while (i < S.Length)
            {
                Point p1 = K.Pop();
                Point p2 = Point.Empty;
                if(K.Count > 0)
                    p2 = K.Peek();
                
                // p2 not on stack or S[i] is left of P2->P1
                if ( p2.IsEmpty || isPLeftOfVector(S[i], p2, p1) )
                {
                    K.Push(p1);
                    K.Push(S[i]);
                    i++;
                }
            }

            // create result array from the stack
            Point[] result = new Point[K.Count];
            for (int ii = 0; ii < result.Length; ii++)
                result[result.Length-ii-1] = K.Pop();

            return result;
        }

        /// <summary>
        /// Returns true, if the point p is on the left hand side of the vector from->to.
        /// </summary>
        private bool isPLeftOfVector(Point p, Point from, Point to)
        {
            return (((to.X - from.X) * (p.Y - from.Y) - (p.X - from.X) * (to.Y - from.Y)) < 0);
        }
    }
}
